@@ -23,8 +23,10 @@ import re
 import locale
 import glob
 import StringIO
-import zlib
 import gzip
+import shutil
+import sys
+import traceback
 
 
 month_names = {"Jan": 1, "Feb": 2, "Mar": 3, "Apr": 4, "May": 5, "Jun": 6,
@@ -57,7 +59,42 @@ item_comment_prefix = "feed item: "
 print_version_keywords_default = [u"printable version", u"версия для печати"]
 single_page_keywords_default = [u"single page", u"одной страницей"]
 next_page_keywords_default = [u"next page"]
-cached_comment_prefix = "cached item: "
+
+try:
+    import html5lib
+    from html5lib import treebuilders
+
+    def parse_html(data = None, fromEncoding = None):
+        if data:
+            parser = html5lib.HTMLParser(tree=treebuilders.getTreeBuilder("beautifulsoup"))
+            return parser.parse(data, encoding = fromEncoding)
+        return BeautifulSoup()
+
+
+    def parse_html_fragment(data, fromEncoding = None):
+        parser = html5lib.HTMLParser(tree=treebuilders.getTreeBuilder("beautifulsoup"))
+        return parser.parseFragment(data)
+
+
+#    def parse_xml(data):
+#        parser = html5lib.XMLParser(tree=treebuilders.getTreeBuilder("beautifulsoup"))
+#        return parser.parse(data)
+    def parse_xml(data):
+        return BeautifulStoneSoup(data, markupMassage = massage)
+
+except ImportError:
+    def parse_html(data = None, fromEncoding = None):
+        if data:
+            return BeautifulSoup(data, fromEncoding = fromEncoding)
+        return BeautifulSoup()
+
+
+    def parse_html_fragment(data, fromEncoding = None):
+        return parse_html(data, fromEncoding = fromEncoding)
+
+
+    def parse_xml(data):
+        return BeautifulStoneSoup(data, markupMassage = massage)
 
 
 def parse_date(str):
@@ -240,30 +277,6 @@ def get_page_base(page, html):
     return page.geturl()
 
 
-def get_page_data(url, **kwargs):
-    if not "headers" in kwargs:
-        kwargs["headers"] = {}
-    kwargs["headers"]["Accept-Encoding"] = "gzip,deflate"
-    page = urlopener.open(urllib2.Request(url=quote_url_path(url), **kwargs))
-    content = page.read()
-    data = content
-    if "Content-Encoding" in page.info():
-        enc = page.info()["Content-Encoding"]
-        if enc == "deflate":
-            data = zlib.decompress(content)
-        elif enc == "gzip":
-            sio = StringIO.StringIO(content)
-            try:
-                gzf = gzip.GzipFile(fileobj = sio)
-                try:
-                    data = gzf.read()
-                finally:
-                    gzf.close()
-            finally:
-                sio.close()
-    return (page, data)
-
-
 def get_body_or_html(html):
     if html and html.body:
         return html.body
@@ -286,11 +299,11 @@ def get_feed_details(url, icon_url = None, page = None):
     if not page:
         try:
             page = urlopener.open(url)
-            xml = BeautifulStoneSoup(page, markupMassage = massage)
+            xml = parse_xml(page)
         finally:
             page.close()
     else:
-        xml = BeautifulStoneSoup(page, markupMassage = massage)
+        xml = parse_xml(page)
     title = get_feed_title(xml)
     icon_tag = xml.find(["icon", "logo", "image"])
     if icon_tag:
@@ -310,7 +323,7 @@ def get_feed_details(url, icon_url = None, page = None):
         if html_url:
             html_page = urlopener.open(html_url)
             try:
-                html = BeautifulSoup(html_page, fromEncoding = get_page_encoding(html_page))
+                html = parse_html(html_page, get_page_encoding(html_page))
                 base = get_page_base(html_page, html)
             finally:
                 html_page.close()
@@ -340,10 +353,11 @@ def discover_feed(url):
     try:
         icon_url = None
         content_type = None
+        content = page.read()
         if page.info().has_key("Content-Type"):
             content_type = page.info()["Content-Type"]
         if content_type and content_type.find("html") >=0:
-            html = BeautifulSoup(page, fromEncoding = get_page_encoding(page))
+            html = parse_html(content, get_page_encoding(page))
             base = get_page_base(page, html)
             icon_tag = html.find(name = "link", attrs = {"rel": re.compile("icon", flags = re.I)})
             if icon_tag:
@@ -353,12 +367,54 @@ def discover_feed(url):
                 rss_tag = html.find(name = "link", attrs = {"rel": "alternate", "type": "application/atom+xml"})
             if rss_tag:
                 return get_feed_details(urljoin(base, rss_tag["href"]), icon_url)
-        return get_feed_details(url, icon_url, page)
+        return get_feed_details(url, icon_url, content)
     finally:
         page.close()
 
 
-class Downloader:
+class PageDownloader(Thread):
+
+
+    def __init__(self, callback, url, **kwargs):
+        Thread.__init__(self)
+        self.daemon = True
+        self.__callback = callback
+        self.__url = url
+        self.__kwargs = kwargs 
+        self.__result = None
+
+
+    def run(self):
+        if not "headers" in self.__kwargs:
+            self.__kwargs["headers"] = {}
+        self.__kwargs["headers"]["Accept-Encoding"] = "gzip"
+        page = urlopener.open(urllib2.Request(url=quote_url_path(self.__url), **self.__kwargs))
+        content = page.read()
+        data = content
+        if "Content-Encoding" in page.info():
+            if page.info()["Content-Encoding"] == "gzip":
+                sio = StringIO.StringIO(content)
+                try:
+                    gzf = gzip.GzipFile(fileobj = sio)
+                    try:
+                        data = gzf.read()
+                    finally:
+                        gzf.close()
+                finally:
+                    sio.close()
+        self.__result = (page, data)
+
+
+    def download(self):
+        self.start()
+        while self.isAlive():
+            if self.__callback:
+                self.__callback()
+            self.join(.25)
+        return self.__result
+
+
+class DownloadQueue:
     """ This class helps managing multithreading downloads. """
 
     def __init__(self, callback):
@@ -374,6 +430,7 @@ class Downloader:
         path_for_url = os.path.join(rel_path, filename)
         self.__cached[url] = path_for_url
         thread = Thread(target = self.__retrieve, args = (quote_url_path(url), os.path.join(abs_path, filename)))
+        thread.daemon = True
         self.__threads.append(thread)
         thread.start()
         return path_for_url
@@ -384,12 +441,15 @@ class Downloader:
             urllib.urlretrieve(url, filename)
         except Exception, ex:
             if self.__callback:
-                self.__callback(("error", ex, None))
+                self.__callback(exception = ex)
 
 
     def wait_completion(self):
         for thread in self.__threads:
-            thread.join()
+            while thread.isAlive():
+                if self.__callback:
+                    self.__callback()
+                thread.join(.25)
 
 
 class Item:
@@ -404,7 +464,7 @@ class Item:
         single_page_keywords = single_page_keywords_default,
         next_page_keywords = next_page_keywords_default,
         page_template = None, allow_scripts = False, content_kwargs = None,
-        reformatter_url = None, downloader = None, ignore_images = False):
+        reformatter_url = None, download_queue = None, ignore_images = False):
         """ It is clear that this is constructor :)
 
         Item is represented either by its content, when tag is specified, or by reference
@@ -450,14 +510,19 @@ class Item:
         self.__content_kwargs = content_kwargs
         self.__reformatter_url = reformatter_url
         self.__next_page_a = None
-        self.__downloader = downloader
-        if not self.__downloader:
-            self.__downloader = Downloader(callback)
+        self.__download_queue = download_queue
+        if not self.__download_queue:
+            self.__download_queue = DownloadQueue(self.__report)
 
 
-    def __report(self, action, item):
+    def __report(self, action = None, item = None, exception = None):
+        details = None
+        if exception:
+            action = "error"
+            item = str(exception)
+            details = ''.join(traceback.format_tb(sys.exc_info()[2]))
         if self.__callback:
-            self.__callback((action, item, None))
+            self.__callback((action, item, None, details))
 
 
     def __calc_the_tag(self, tag):
@@ -486,7 +551,7 @@ class Item:
     def __get_cache_directory(self):
         cache_path = os.path.join(self.__path, self.__cache_path)
         if not os.path.exists(cache_path):
-            os.mkdir(cache_path)
+            os.makedirs(cache_path)
         return cache_path
 
 
@@ -496,10 +561,11 @@ class Item:
         kwargs are passed to Request object constructor.
         """
         self.__report("downloading", url)
-        (page, data) = get_page_data(url, **kwargs)
+        downloader = PageDownloader(self.__report, url, **kwargs)
+        (page, data) = downloader.download()
         try:
             self.__report("parsing", url)
-            html = BeautifulSoup(data, fromEncoding = get_page_encoding(page))
+            html = parse_html(data, get_page_encoding(page))
             base = get_page_base(page, html)
             url = page.geturl()
         finally:
@@ -522,7 +588,7 @@ class Item:
                     a["href"] = item.save()
                     return a
         except Exception, ex:
-            self.__report("error", ex)
+            self.__report(exception = ex)
 
 
     def __find_single_page(self, html, base, referer):
@@ -535,7 +601,7 @@ class Item:
                     (new_html, new_base, real_url) = self.__get_page_html(url, headers={"Referer": referer})
                     return (new_html, new_base, True)
         except Exception, ex:
-            self.__report("error", ex)
+            self.__report(exception = ex)
         return (html, base, False)
 
 
@@ -549,7 +615,7 @@ class Item:
                     (new_html, new_base, real_url) = self.__get_page_html(url, headers={"Referer": referer})
                     return (new_html, new_base, get_body_or_html(new_html))
         except Exception, ex:
-            self.__report("error", ex)
+            self.__report(exception = ex)
         return (html, base, None)
 
 
@@ -602,12 +668,10 @@ class Item:
         if tag.has_key(attr):
             try:
                 url = urljoin(base, tag[attr])
-                path_for_url = self.__downloader.queue(url, 
+                tag[attr] = self.__download_queue.queue(url, 
                     self.__get_cache_directory(), self.__cache_path)
-                tag.parent.append(Comment(NavigableString(cached_comment_prefix + path_for_url)))
-                tag[attr] = path_for_url
             except Exception, ex:
-                self.__report("error", ex)
+                self.__report(exception = ex)
 
 
     def __cache_urls(self, tags, attr, base):
@@ -631,17 +695,15 @@ class Item:
             content = [self.__tag.findParent("table")]
         else:
             content = [node for node in self.__tag.contents]
-        if self.__next_page_a:
-            content.append(Comment(NavigableString(cached_comment_prefix + self.__next_page_a["href"])))
-            if not tag_contains(self.__tag, self.__next_page_a):
-                self.__next_page_a.extract()
-                content.append(self.__next_page_a)
+        if self.__next_page_a and not tag_contains(self.__tag, self.__next_page_a):
+            self.__next_page_a.extract()
+            content.append(self.__next_page_a)
         return content
 
 
     def save(self, title = None, filename = None):
         """ Saves item into cache. """
-        html = BeautifulSoup(self.__page_template)
+        html = parse_html(self.__page_template)
         if not title:
             title = self.title
         get_or_create_tag(html, ["head", "title"]).append(NavigableString(title))
@@ -649,6 +711,8 @@ class Item:
         [root.append(tag) for tag in self.__contents]
         if not filename:
             filename = uuid.uuid1().hex + ".html"
+        if not os.path.exists(self.__path):
+            os.makedirs(self.__path)
         f = open(os.path.join(self.__path, filename), "w")
         f.write(html.prettify())
         f.close()
@@ -658,43 +722,12 @@ class Item:
     def process(self):
         """ Entry point. """
         self.__contents = []
-        try:
-            if self.__url:
-                self.__extract_tag()
-            self.__contents = self.__fixup_tag()
-        except BaseException, ex:
-            self.__downloader.wait_completion()
-            self.delete()
-            raise
+        if self.__url:
+            self.__extract_tag()
+        self.__contents = self.__fixup_tag()
         return self.__contents
 
 
-    def __delete_cached_item(self, filename):
-        if os.path.exists(filename):
-            if filename.endswith(".html"):
-                f = open(filename, "r")
-                html = BeautifulSoup(f)
-                f.close()
-                self.__delete_cached_items(html, os.path.dirname(filename))
-            self.__report("deleting", filename)
-            os.remove(filename)
-
-
-    def __delete_cached_items(self, tag, path):
-        items_to_delete = tag.findAll(text = lambda text:
-            isinstance(text, Comment) and text.startswith(cached_comment_prefix))
-        [self.__delete_cached_item(os.path.join(path, item[len(cached_comment_prefix):]))
-            for item in items_to_delete]
-
-
-    def delete(self, filename = None):
-        """ Deletes item and all related resources. """
-        if filename:
-            self.__delete_cached_item(os.path.join(self.__path, filename))
-        elif self.__tag:
-            self.__delete_cached_items(self.__tag, self.__path)
-
-
 class Feed:
     """ Collection of feed related operations. """
     
@@ -703,7 +736,7 @@ class Feed:
         print_version_keywords = print_version_keywords_default,
         single_page_keywords = single_page_keywords_default,
         next_page_keywords = next_page_keywords_default,
-        delete_before_date = None, allow_scripts = False, inline_items = False, cache_items = False,
+        days_to_keep = None, allow_scripts = False, inline_items = False, cache_items = False,
         include = None, exclude = None, content_kwargs = None, reformatter_url = None,
         new_window = False, always_clear_cache = False, ignore_images = False):
         """ Constructor.
@@ -718,7 +751,7 @@ class Feed:
         feed_item_template - html used to construct feed item.
         print_version_keywords - a list of phrases to scan html document searching for reference to
         printable version.
-        delete_before_date - all items before this date are deleted during feed update.
+        days_to_keep - days to keep items before delete them.
         allow_scripts - do not remove script tags from item before caching.
         inline_items - inline referenced resources into feed html.
         cache_items - download and cache referenced resources.
@@ -756,7 +789,8 @@ class Feed:
         self.__next_page_keywords = next_page_keywords
         self.__inline_items = inline_items
         self.__cache_items = cache_items
-        self.__delete_before_date = delete_before_date
+        if days_to_keep:
+            self.__delete_before_date = date.today() - timedelta(days_to_keep)
         self.__allow_scripts = allow_scripts
         self.__ignore_images = ignore_images
         self.__include = None
@@ -769,12 +803,18 @@ class Feed:
         self.__reformatter_url = reformatter_url
         self.__new_window = new_window
         self.__always_clear_cache = always_clear_cache
-        self.__downloader = Downloader(callback)
+        self.__download_queue = DownloadQueue(self.__report)
+        self.new_item_count = 0
 
 
-    def __report(self, action, item, percent = None):
+    def __report(self, action = None, item = None, percent = None, exception = None):
+        details = None
+        if exception:
+            action = "error"
+            item = str(exception)
+            details = ''.join(traceback.format_tb(sys.exc_info()[2]))
         if self.__callback:
-            self.__callback((action, item, percent))
+            self.__callback((action, item, percent, details))
 
 
     def __unquote_special(self, str):
@@ -783,6 +823,17 @@ class Feed:
         return str
 
 
+    def __get_item_title(self, item):
+        if item.title.string:
+            return self.__unquote_special(item.title.string.strip())
+        for node in item.title:
+            if node.string:
+                str = node.string.strip()
+                if str:
+                    return str
+        return ""
+
+
     def __get_item_link(self, item):
         if item.guid:
             if not item.guid.has_key("ispermalink") or item.guid["ispermalink"].lower() == "true":
@@ -790,8 +841,15 @@ class Feed:
         if item.link:
             if item.link.has_key("href"):
                 return item.link["href"] #Atom uses href to store links
-            else:
+            elif item.link.string:
                 return item.link.string
+            else:
+                # Once there was a feed with url in CData section
+                for node in item.link:
+                    if node.string:
+                        str = node.string.strip()
+                        if str:
+                            return str
 
 
     def __get_item_date(self, item):
@@ -812,8 +870,8 @@ class Feed:
             if str:
                 if not isinstance(str, CData):
                     str = self.__unquote_special(str)
-                return BeautifulSoup(str)
-        return BeautifulSoup()
+                return parse_html_fragment(str)
+        return parse_html()
 
 
     def __item_exists(self, tag, link):
@@ -850,22 +908,23 @@ class Feed:
             next_page_keywords = self.__next_page_keywords,
             allow_scripts = self.__allow_scripts, content_kwargs = self.__content_kwargs,
             reformatter_url = self.__reformatter_url,
-            downloader = self.__downloader, page_template = self.__page_template, ignore_images = ignore_images)
+            download_queue = self.__download_queue, page_template = self.__page_template, ignore_images = ignore_images)
         
 
     def __get_update_html(self, html):
         div = Tag(html, "div")
         div["class"] = "feedcircuit-update"
-        div["id"] = datetime.today().strftime("%Y%m%d%H%M%S")
+        now = datetime.today()
+        div["id"] = now.strftime("%Y%m%d%H%M%S") + "." + str(now.microsecond)
         template = self.__feed_update_template.replace("%title", self.__title.strip())
         template = template.replace("%timestamp",
             unicode(datetime.today().strftime("%c").decode(locale.getpreferredencoding())))
-        div.append(BeautifulSoup(template))
+        div.append(parse_html_fragment(template))
         return div
 
 
     def __get_item_html(self):
-        template = BeautifulSoup(self.__feed_item_template)
+        template = parse_html_fragment(self.__feed_item_template)
         item = Tag(template, "div")
         item["class"] = "feedcircuit-item"
         item.append(template)
@@ -889,34 +948,37 @@ class Feed:
     def __add_item(self, div, item, percent):
         """ Here Item is get included to feed after executing all dependant operations. """
         link = self.__get_item_link(item)
-        title = self.__unquote_special(item.title.string.strip())
+        title = self.__get_item_title(item)
         self.__report("processing", title, percent)
+        cache_path = self.__cache_path
+        if div.has_key("id"):
+            cache_path = os.path.join(cache_path, div["id"])
         (item_html, item_a, item_div) = self.__get_item_html()
-        div.append(Comment(NavigableString(item_comment_prefix + link)))
-        div.append(item_html)
         item_a["href"] = link
         item_a.append(NavigableString(title))
         if self.__inline_items:
-            feed_item = self.__create_feed_item(url = link, ignore_images = self.__ignore_images)
+            feed_item = self.__create_feed_item(url = link, ignore_images = self.__ignore_images, cache_path = cache_path)
             [item_div.append(t) for t in feed_item.process()]
         else:
             feed_item = self.__create_feed_item(tag = self.__get_item_description(item),
-                base = self._real_url, ignore_images = self.__ignore_images)
+                base = self._real_url, ignore_images = self.__ignore_images, cache_path = cache_path)
             [item_div.append(t) for t in feed_item.process()]
             if self.__cache_items:
-                path = os.path.join(self.__path, self.__cache_path)
+                path = os.path.join(self.__path, cache_path)
                 feed_item = self.__create_feed_item(url = link, path = path, cache_path = "")
                 feed_item.process()
-                item_a["href"] = os.path.join(self.__cache_path, feed_item.save(title))
+                item_a["href"] = os.path.join(cache_path, feed_item.save(title))
                 if self.__new_window:
                     item_a["target"] = "_blank"
-                item_div.insert(0, Comment(NavigableString(cached_comment_prefix + item_a["href"])))
-
+        div.append(Comment(NavigableString(item_comment_prefix + link)))
+        div.append(item_html)
 
 
     def __delete_div(self, div):
-        feed_item = self.__create_feed_item(tag = div)
-        feed_item.delete()
+        if div.has_key("id"):
+            div_path = os.path.join(os.path.join(self.__path, self.__cache_path), div["id"])
+            if os.path.exists(div_path):
+                shutil.rmtree(div_path, True)
         div.extract()
 
 
@@ -928,14 +990,15 @@ class Feed:
     def __delete_before(self, html):
         if self.__delete_before_date:
             divs = filter(lambda div:
-                div.has_key("id") and datetime.strptime(div["id"], "%Y%m%d%H%M%S").date() < self.__delete_before_date,
+                div.has_key("id") and datetime.strptime(div["id"][:14], "%Y%m%d%H%M%S").date() < self.__delete_before_date,
                 get_feedcircuit_root(html).findAll("div", recursive = False))
             [self.__delete_div(div) for div in divs]
 
 
-    def __get_new_item_count(self, tag):
-        return len(tag.findAll(text = lambda text: isinstance(text, Comment)
-            and text.startswith(item_comment_prefix)))
+    def __save_html(self, path, html):
+        f = open(path, "w")
+        f.write(html.prettify())
+        f.close()
 
 
     def update(self):
@@ -944,15 +1007,16 @@ class Feed:
         When there are now new items feed file stays untouched.
         Outdated items are deleted only when there are some new ones.
         """
-        result = 0
+        self.new_item_count = 0
         div = None
         try:
             self.__report("downloading", self.__url, 1)
-            (page, data) = get_page_data(self.__url)
+            downloader = PageDownloader(self.__report, self.__url)
+            (page, data) = downloader.download()
             try:
                 self._real_url = page.geturl()
                 self.__report("parsing", self.__url, 10)
-                xml = BeautifulStoneSoup(data, markupMassage = massage)
+                xml = parse_xml(data)
             finally:
                 page.close()
             if not self.__title:
@@ -962,13 +1026,13 @@ class Feed:
                 path = os.path.join(self.__path, self.__filename)
             if os.path.exists(path):
                 f = open(path, "r")
-                html = BeautifulSoup(f)
+                html = parse_html(f)
                 if self.__always_clear_cache:
                     self.__report("deleting", "", 15)
                     self.__delete_all_divs(html)
                 f.close()
             else:
-                html = BeautifulSoup(self.__feed_template)
+                html = parse_html(self.__feed_template)
                 get_or_create_tag(html, ["head", "title"]).append(NavigableString(self.__title))
             self.__existing_items = html.findAll(text = lambda text:
                 isinstance(text, Comment) and text.startswith(item_comment_prefix))
@@ -981,28 +1045,25 @@ class Feed:
                     and not self.__item_exists(div.parent, link):
                     to_process.append(item)
             count = len(to_process)
+            if count:
+                self.__save_html(path, html)
             for i in range(count):
                 try:
                     self.__add_item(div, to_process[i], 20 + (i * 60 / count))
+                    self.new_item_count += 1
                 except Exception, ex:
-                    self.__report("error", ex)
-            result = self.__get_new_item_count(div)
-            if result:
-                if not self.__always_clear_cache:
-                    self.__report("deleting", "", 80)
-                    self.__delete_before(html)
-                self.__report("downloading images", "", 90)
-                self.__downloader.wait_completion()
-                f = open(path, "w")
-                f.write(html.prettify())
-                f.close()
+                    self.__report(exception = ex)
+            self.__report("downloading images", "", 90)
+            self.__download_queue.wait_completion()
         except BaseException, ex:
-            self.__report("error", ex)
-            self.__downloader.wait_completion()
-            if div:
-                self.__delete_div(div)
+            self.__report(exception = ex)
             raise
-        return result
+        finally:
+            if self.new_item_count:
+                if not self.__always_clear_cache:
+                    self.__report("deleting", "", 95)
+                    self.__delete_before(html)
+                self.__save_html(path, html)
 
 
     def delete(self):
@@ -1013,7 +1074,7 @@ class Feed:
             path = os.path.join(self.__path, self.__title + ".html")
         if os.path.exists(path):
             f = open(path, "r")
-            html = BeautifulSoup(f)
+            html = parse_html(f)
             f.close()
             self.__delete_all_divs(html)
             os.remove(path)
